/*
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License along
 with this program; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/*
 * This is a midi Sequence, or a "file". Each tab in the tab bar represents one Sequence instance.
 * It contains general information and a vector of tracks.
 */

#include "Config.h"

#include "GUI/GLPane.h"
#include "GUI/MainFrame.h"
#include "GUI/MeasureBar.h"

#include "Midi/Players/PlatformMidiManager.h"
#include "Midi/Sequence.h"
#include "Midi/Track.h"

#include "Midi/CommonMidiUtils.h"
#include "IO/IOUtils.h"

#include "OpenGL.h"

#include "irrXML/irrXML.h"

namespace AriaMaestosa {
	

Sequence::Sequence(GLPane* parent)
{
    INIT_LEAK_CHECK();
	
    glPane = parent;
    scrollToPlaybackPosition=false;
    
    reordering_newPosition = -1;
    
    mainFrame = glPane->mainFrame;
    
	somethingDoneToUndo = false;
	
    dockHeight=0;
    
    beatResolution = 960;
    
    dockSize=0;
    currentTrack=0;
    tempo=120;
    x_scroll_in_pixels = 0;
    y_scroll = 0;
    reorderYScroll = 0;
	
    startTick=0; // from the song starts - this is only to know where the progress bar should start
    
    importing=false;
    
    measureAmount=DEFAULT_SONG_LENGTH;
    firstMeasure=0;
    
	x_scroll_upon_copying = -1;
	
	follow_playback = false;
	
    measureBottomValue = 4;
    measureTopValue = 4;
    
	setZoom(100);
	
    sequenceName = wxGetTranslation("Untitled song");
    
    if(glPane->isVisible) addTrack();
    
    copyright = "";
    
}

Sequence::~Sequence()
{
	tracks.clearAndDeleteAll();
}

// ------------------------------------------- copyright ----------------------------------------
void Sequence::setCopyright( char* copyright )
{
    Sequence::copyright = std::string(copyright);
}

std::string Sequence::getCopyright()
{
    return copyright;
}


// ------------------------------------------- grid ----------------------------------------

/*
 * When user selects 'snap notes to grid' from menu. Simply disptaches the event to the appropriate track.
 */

void Sequence::snapNotesToGrid()
{
    assert(currentTrack>=0);
    assert(currentTrack<tracks.size());
    
    tracks[ currentTrack ].snapNotesToGrid();
    
    glPane->render();
}

// ------------------------------------------- scrolling ----------------------------------------

int Sequence::getYScroll()
{
    return y_scroll;   
}

int Sequence::getXScrollInMidiTicks()
{
    return (int)( x_scroll_in_pixels/zoom );
}

int Sequence::getXScrollInPixels()
{
    return x_scroll_in_pixels;
}

void Sequence::setXScrollInMidiTicks(int value)
{
	x_scroll_in_pixels = (int)( value * zoom );
    
    glPane->render();
}
void Sequence::setXScrollInPixels(int value)
{
    x_scroll_in_pixels = value;
    
    glPane->render();
}

void Sequence::setYScroll(int value)
{
    y_scroll = value;
}

// ------------------------------------------- zoom ----------------------------------------

float Sequence::getZoom()
{
    return zoom;
}

int Sequence::getZoomInPercent()
{
	return zoom_percent;
}

void Sequence::setZoom(int zoom)
{
    Sequence::zoom = (128.0/(float)ticksPerMeasure())*(zoom/100.0);
	Sequence::zoom_percent = zoom;
}

// ------------------------------------------- undo ----------------------------------------

/*
 * Called just before doing any operation that modifies notes.
 * Tells the appropriate track to do a copy of current state, to be able to restore it if user wanted to.
 */

void Sequence::loadUndoMemory()
{
	if(!somethingDoneToUndo) return;
    tracks[lastModifiedTrack].loadUndoMemory();
    mainFrame->glPane->render();
}


// ------------------------------------------- render ----------------------------------------

void Sequence::renderTracks(int currentTick, int mousex, int mousey, int mousey_initial, int from_y)
{
    
	const int draggedTrack = glPane->getDraggedTrackID();

	// draw tracks normally
    if(draggedTrack==-1)
	{
        reorderYScroll=0;
        
        int y=from_y - y_scroll;
        const int trackAmount = tracks.size();
        for(int n=0; n<trackAmount; n++)
		{
            tracks[n].setId(n);
            y = tracks[n].graphics->render(y, currentTick, n==currentTrack);
        }

    }
	// reordering
	else
	{

		
        reordering_newPosition = draggedTrack;
        
        // draw tracks before current
		int first_y = mousey_initial - reorderYScroll;
		
        for(int tracknum=draggedTrack-1; tracknum>=0; tracknum--)
		{
			if(tracks[tracknum].graphics->docked) continue;
			
			first_y =  mousey_initial - (draggedTrack-tracknum)*50 - reorderYScroll;
            tracks[tracknum].graphics->renderHeader(0, first_y, true);
            if(mousey < mousey_initial - (draggedTrack-tracknum)*50+25 - reorderYScroll) reordering_newPosition=tracknum;
        }
        
        // draw tracks after current
		int last_y = mousey_initial - reorderYScroll;
		
        for(int tracknum=draggedTrack+1; tracknum<tracks.size(); tracknum++)
		{
			if(tracks[tracknum].graphics->docked) continue;
			
			last_y = mousey_initial + (tracknum-draggedTrack)*50 - reorderYScroll;
            tracks[tracknum].graphics->renderHeader(0, last_y, true);
            if(mousey > mousey_initial + (tracknum-draggedTrack)*50+25 - reorderYScroll) reordering_newPosition=tracknum+1;
        }
        
		// scroll up or down if mouse goes to the edges
		if(mousey < 100)
		{
			if(mousey > first_y)
				reorderYScroll -= (100-mousey)*4/100;
		}
		
        if(mousey > glPane->getHeight()-100)
		{
			if(mousey < last_y+50)
				reorderYScroll += (mousey - glPane->getHeight()+100)*4/100;
        }
		
        // draw track the user is dragging
        tracks[draggedTrack].graphics->renderHeader(mousex, mousey/* - reorderYScroll*/, true);

        const int arrow_y = (reordering_newPosition > draggedTrack) ?
            mousey_initial + (reordering_newPosition - draggedTrack)*50 - 5 - reorderYScroll :
            mousey_initial - (draggedTrack - reordering_newPosition)*50 - 5 - reorderYScroll;
           
        glColor3f(1,0,0);
        glBegin(GL_LINES);
        glVertex2f( 26, arrow_y);
        glVertex2f( 10, arrow_y);
        glEnd();
        
        glBegin(GL_TRIANGLE_STRIP);
        glVertex2f( 35, arrow_y);
        glVertex2f( 25, arrow_y - 5);
        glVertex2f( 25, arrow_y + 5);
        glEnd();
    }//end if
    
}



// ------------------------------------------- mouse events ----------------------------------------

/*
 * do we need to start a timer that will frequently send mouse held down events?
 */

bool Sequence::areMouseHeldDownEventsNeeded()
{
	
	const int draggedTrack = glPane->getDraggedTrackID();
	
	// we're reordering tracks, it is necessary. return true.
	if(draggedTrack!=-1) return true;
	
	// ask all editors if they need such events at this point.
	const int trackAmount = tracks.size();
	for(int n=0; n<trackAmount; n++)
	{
		// there's one editor that wants them, so return true.
		if(tracks[n].graphics->getCurrentEditor()->areMouseHeldDownEventsNeeded()) return true;
	}//next
	
	// we're not reordering and no editor requested such events, so return false
	return false;
}

/*
 *  Called repeatedly when mouse is held down
 */

void Sequence::mouseHeldDown(int mousex_initial, int mousey_initial, int mousex_current, int mousey_current)
{

	const int draggedTrack = glPane->getDraggedTrackID();

	// if reordering tracks
	if(draggedTrack!=-1)
	{
		glPane->render(); // reordering preview is done while rendering, so calling 'render' will update reordering onscreen.
		return;
	}
	
	// dispatch event to all tracks
	const int trackAmount = tracks.size();
	for(int n=0; n<trackAmount; n++)
	{
		tracks[n].graphics->getCurrentEditor()->mouseHeldDown();
	}//next
	
}



// ------------------------------------------- resolution ----------------------------------------

/*
 * Get various information about resolution.
 * Ticks per beat is the number of time units in a quarter note.
 * Ticks per measure returns how many ticks there would be in a whole measure considering the beat resolution and the measure time signature.
 */

int Sequence::ticksPerBeat(){		return beatResolution;		}

int Sequence::ticksPerMeasure()
{
	return (int)(
				 beatResolution * measureTopValue * (4.0/measureBottomValue)
				 );

}

void Sequence::setTicksPerBeat(int res){		beatResolution = res;		}

// ------------------------------------------- info ----------------------------------------

/*
 * Returns the number of pixels it takes to draw all tracks, vertically.
 * This is used mostly by the code managing the vertical scrollbar.
 */

int Sequence::getTotalHeight()
{
    
    int totalHeight=0;
    
    for(int n=0; n<tracks.size(); n++){
        totalHeight += tracks[n].graphics->getTotalHeight() + 10;    
    }
    
    return totalHeight;
}

// ------------------------------------------- tracks ----------------------------------------

void Sequence::addTrack()
{
    //tracks.push_back(new Track(mainFrame, this));
	if(currentTrack>=0 and currentTrack<tracks.size()) tracks.add(new Track(mainFrame, this), currentTrack+1); // add new track below active one
	else tracks.push_back(new Track(mainFrame, this));
	
    glPane->render();
}

void Sequence::deleteTrack()
{
    if(currentTrack<0 or currentTrack>tracks.size()-1) return;
    
    tracks.erase( currentTrack );
    glPane->render();
}

void Sequence::deleteTrack(int ID)
{
	assertExpr(ID,>=,0);
	assertExpr(ID,<,tracks.size());
	
	tracks.erase( ID );
}

/*
 * Called when a user has finished dragging the track to reorder it.
 * Where the track ends was calculated while drawing the preview - all this methods needs to do is
 * remove the track from its curren location and move it to its new location.
 */

void Sequence::reorderTracks()
{
	
    const int draggedTrack = glPane->getDraggedTrackID();
	
    if( reordering_newPosition == draggedTrack ) return;
    if( reordering_newPosition == -1 ) return;
    
    Track* dragged_track = &tracks[draggedTrack];
	
    // if we remove an element now, the IDs in the vector will be modified and reordering_newPosition may not be valid anymore.
	// so we just 'mark' the element we don't want anymore, and it will be removed just after adding the new item.
    tracks.markToBeRemoved(draggedTrack);
    tracks.add(dragged_track, reordering_newPosition);
    tracks.removeMarked();
    
    currentTrack = reordering_newPosition-1;

	if(! (currentTrack>=0) ) currentTrack=0;
	if(! (currentTrack<tracks.size()) ) currentTrack=0;
}

/*
 * Hide a track by sending it to the 'dock'
 */

void Sequence::addToDock(GraphicalTrack* track)
{
    
    dock.push_back(track);
    dockSize = dock.size();
    
	currentTrack = 0;
	
    mainFrame->updateVerticalScrollbar();
}

void Sequence::removeFromDock(GraphicalTrack* track)
{
    
    for(int n=0; n<dock.size(); n++)
	{
        if(&dock[n] == track)
		{
            dock.remove( n );
            dockSize = dock.size();
            return;
        }
    }
    mainFrame->updateVerticalScrollbar();
    
}

Track* Sequence::getCurrentTrack()
{
	return &tracks[currentTrack];	
}

int Sequence::getTrackAmount()
{
	return tracks.size();
}

int Sequence::getCurrentTrackID()
{
	return currentTrack;
}

void Sequence::setCurrentTrackID(int ID)
{
	currentTrack = ID;
}

Track* Sequence::getTrack(int ID)
{
	assertExpr(ID,>=,0);
	assertExpr(ID,<,tracks.size());
	
	return &tracks[ID];
}

void Sequence::setCurrentTrack(Track* track)
{
	const int trackAmount = tracks.size();
	for(int n=0; n<trackAmount; n++)
	{
		if( &tracks[n] == track )
		{
			currentTrack = n;
			return;
		}
	}
	
	std::cout << "Error: void Sequence::setCurrentTrack(Track* track) couldn't find any matching track" << std::endl;
	
}

// ------------------------------------------- playback ----------------------------------------

void Sequence::scrollNowToPlaybackPosition(){		scrollToPlaybackPosition=true;		}

void Sequence::enterPlayLoop()
{

    int width_tick =  ticksPerMeasure();

    int lastTick = -1;
    while( true )
	{
		
        const int currentTick = PlatformMidiManager::trackPlaybackProgression();
        if(currentTick == -1)
		{
            break; // song is over
        }
        
        if(!PlatformMidiManager::isPlaying()) break;
        
        if(lastTick != startTick + currentTick)
		{ // only draw if it has changed
            
            // if user has clicked on a little red arrow
            if(scrollToPlaybackPosition)
			{
                scrollToPlaybackPosition=false;
                x_scroll_in_pixels = (int)( (startTick + currentTick)*zoom );
                mainFrame->setHorizontalScrollbar( startTick + currentTick );
            }
            
			// if follow playback is checked in the menu
			if(follow_playback)
			{
				x_scroll_in_pixels = (int)( (startTick + currentTick - width_tick)*zoom );
				if( x_scroll_in_pixels < 0 ) x_scroll_in_pixels = 0;
				mainFrame->setHorizontalScrollbar( startTick + currentTick - ticksPerMeasure() );
			}
			
            glPane->setCurrentTick( startTick + currentTick );
            glPane->render();
            lastTick = startTick + currentTick;
        }

        wxYield();
        wxMilliSleep(10);
    }// wend
    
    
    glPane->setCurrentTick( -1 );
    glPane->render();

}

void Sequence::spacePressed()
{

    if(!PlatformMidiManager::isPlaying())
	{

        if( mainFrame->playback_mode )
		{
			return;
		}
		
        mainFrame->toolsEnterPlaybackMode();
		
		startTick = -1;
        bool success = PlatformMidiManager::playSelected(this, &startTick);
		
        if(!success or startTick == -1 /*failure*/)
		{
            if(!PlatformMidiManager::isPlaying()) mainFrame->toolsExitPlaybackMode();
			return;
		}
		else
		{
            enterPlayLoop();
		}
        
    }
	else
	{
		
        mainFrame->toolsExitPlaybackMode();
        PlatformMidiManager::stop();
        glPane->render();
        
    }
	
}

// ------------------------------------------- copy/paste ----------------------------------------

void Sequence::copy()
{
    tracks[currentTrack].copy();
	x_scroll_upon_copying = x_scroll_in_pixels;
}

void Sequence::paste()
{
    tracks[currentTrack].paste(false);
    glPane->render();   
}

void Sequence::pasteAtMouse()
{
    tracks[currentTrack].paste(true);
    glPane->render();   
}

// ------------------------------------------- selection ----------------------------------------

void Sequence::selectAll()
{
    tracks[currentTrack].selectNote(ALL_NOTES, true, true);
    glPane->render();
}

void Sequence::selectNone()
{
    tracks[currentTrack].selectNote(ALL_NOTES, false, true);
    glPane->render();
}

// ------------------------------------------- scale ----------------------------------------

void Sequence::scale(
                     float factor,
                     bool rel_first_note, bool rel_begin, // relative to what (only one must be true)
                     bool affect_selection, bool affect_track, bool affect_song // scale what (only one must be true)
                     )
{
    
    
    int relative_to = -1;
    
    // selection
    if(affect_selection)
	{
        
        if(rel_first_note) relative_to = tracks[currentTrack].getFirstNoteTick(true);
        else if(rel_begin) relative_to=0;
     
        tracks[currentTrack].saveUndoMemory();
        tracks[currentTrack].scale(factor, relative_to, true);
    }

    // track
    else if(affect_track)
	{
        
        if(rel_first_note) relative_to = tracks[currentTrack].getFirstNoteTick();
        else if(rel_begin) relative_to=0;
        
        tracks[currentTrack].saveUndoMemory();
        tracks[currentTrack].scale(factor, relative_to, false);
    }
    
    // song
    else if(affect_song)
	{
        
        if(rel_first_note)
		{
            
            // find first tick in all tracks [i.e. find the first tick of all tracks and keep the samllest]
            int song_first_tick = -1;
            for(int n=0; n<tracks.size(); n++)
			{
                
                const int track_first_tick = relative_to = tracks[n].getFirstNoteTick();
                if(track_first_tick<song_first_tick or song_first_tick==-1) song_first_tick = track_first_tick;
            }//next
            
            relative_to = song_first_tick;
        }
        else if(rel_begin) relative_to=0;
        
        // scale all tracks
        for(int n=0; n<tracks.size(); n++)
		{
            tracks[n].scale(factor, relative_to, false);
        }
        
    }

    glPane->render();
    
}

// ------------------------------------------- length ----------------------------------------

void Sequence::setMeasureAmount(int measureAmount)
{
    Sequence::measureAmount=measureAmount;
    glPane->render();
	glPane->measureBar->updateVector(measureAmount);
}

int Sequence::getMeasureAmount()
{
    return measureAmount;
}

// ------------------------------------------- start ----------------------------------------
void Sequence::setFirstMeasure(int firstMeasureID)
{
    firstMeasure = firstMeasureID;   
}

void Sequence::setStartTick(int newValue)
{
    startTick = newValue;
}


// ------------------------------------------- measures ----------------------------------------

void Sequence::setMeasureSig(int top, int bottom)
{
    measureTopValue = top;
	
	float denom = (float)log(bottom)/(float)log(2);
	if( (int)denom != (float)denom )
	{
		wxBell();
		wxMessageBox( wxGetTranslation("Denominator must be a power of 2") );
	}
	
    measureBottomValue = bottom;
    
	setZoom( getZoomInPercent() ); // update zoom to new measure size
	
    wxSpinEvent unused;
    mainFrame->songLengthChanged(unused);
	
    glPane->render();
}

int Sequence::getMeasureTopValue()
{
    return measureTopValue;   
}

int Sequence::getMeasureBottomValue()
{
    return measureBottomValue;
}

// ------------------------------------------- tempo ----------------------------------------

void Sequence::setTempo(int tmp)
{
    tempo = tmp;
}

int Sequence::getTempo()
{
    return tempo;
}



// ------------------------------------------- i/o ----------------------------------------

/*
 * Called before loading, prepares empty tracks
 */

void Sequence::prepareEmptyTracksForLoading(int amount)
{
    
    tracks.clearAndDeleteAll();
    for(int n=0; n<amount; n++) tracks.push_back( new Track(mainFrame, this) );
}

/*
 * Called when saving <section> ... </section> in .aria file
 */

void Sequence::saveToFile(wxFileOutputStream& fileout){
	
	//fileout.WriteString
	
	writeData(wxString("<sequence"), fileout );
	
	writeData(wxString(" maintempo=\"") + to_wxString(tempo) +
						wxString("\" measureAmount=\"") + to_wxString(measureAmount) +
						wxString("\" currentTrack=\"") + to_wxString(currentTrack) +
						wxString("\" beatResolution=\"") + to_wxString(beatResolution) +
						wxString("\">\n\n"), fileout );
	
	writeData(wxString("<view xscroll=\"") + to_wxString(x_scroll_in_pixels) +
						wxString("\" yscroll=\"") + to_wxString(y_scroll) +
						wxString("\" zoom=\"") + to_wxString(zoom_percent) +
						wxString("\"/>\n"), fileout );
	
	writeData(wxString("<measure width=\"") + to_wxString(ticksPerMeasure()) +
						wxString("\" firstMeasure=\"") + to_wxString(firstMeasure) +
						wxString("\" denom=\"") + to_wxString(measureBottomValue) +
						wxString("\" num=\"") + to_wxString(measureTopValue) +
						wxString("\"/>\n\n"), fileout );
	
	writeData(wxString("<tempo>\n"), fileout );
	// tempo changes
    for(int n=0; n<tempoEvents.size(); n++)
	{
        tempoEvents[n].saveToFile(fileout);   
    }
	writeData(wxString("</tempo>\n"), fileout );
	
	writeData(wxString("<copyright>"), fileout );
	writeData(wxString(getCopyright().c_str()), fileout );
	writeData(wxString("</copyright>\n"), fileout );
	
	// tracks
    for(int n=0; n<tracks.size(); n++)
	{
        tracks[n].saveToFile(fileout);   
    }
	
	
	writeData(wxT("</sequence>"), fileout );
	
	somethingDoneToUndo = false;
}



// convert string to double
double string2double(char* digit) {
	int sign = 1;
	
	double result = 0;
	
	// check sign
	if (*digit == '-')
	{
		sign = -1;
		digit++;
	}
	
	// get integer portion
	while (*digit >= '0' && *digit <='9')
	{
		result = (result * 10) + *digit-'0';
		digit++;
	}
	
	// get decimal point and fraction, if any.
	if (*digit == '.' or *digit == ',')
	{
		digit++;
		double scale = 0.1;
		while (*digit >= '0' && *digit <='9')
		{
			result += (*digit-'0') * scale;
			scale *= 0.1;
			digit++;
		}
	}
	
	// error if we're not at the end of the number
	if (*digit != 0) {
		return -1;
	}
	
	//set to sign given at the front
	result = result * sign;
	return result;
}


/*
 * Called when reading <section> ... </section> in .aria file
 */

bool Sequence::readFromFile(irr::io::IrrXMLReader* xml)
{

	tracks.clearAndDeleteAll();
	
	bool copyright_mode = false;
	bool tempo_mode = false;
	
	// parse the file until end reached
    while(xml && xml->read())
    {
        

        switch(xml->getNodeType())
        {
            case irr::io::EXN_TEXT:
				
				if(copyright_mode) setCopyright( (char*)xml->getNodeData() );
				
                break;
            case irr::io::EXN_ELEMENT:
            {
			
				// ---------- sequence ------
                if (!strcmp("sequence", xml->getNodeName()))
				{
                    
                    const char* maintempo = xml->getAttributeValue("maintempo");
					if(maintempo != NULL) tempo = atoi( maintempo );
					else{
						tempo = 120;
						std::cerr << "Missing info from file: main tempo" << std::endl;
					}
					
                    const char* measureAmount_c = xml->getAttributeValue("measureAmount");
					if(measureAmount_c != NULL)
					{
						measureAmount=atoi(measureAmount_c);
						mainFrame->changeMeasureAmount(measureAmount, false);
					}
					else
					{
						std::cerr << "Missing info from file: measure amount" << std::endl;
						return false;
					}
					
					const char* currentTrack_c = xml->getAttributeValue("currentTrack");
					if( currentTrack_c != NULL ) currentTrack = atoi( currentTrack_c );
					else
					{
						currentTrack = 0;
						std::cerr << "Missing info from file: current track" << std::endl;	
					}
					
					const char* beatResolution_c = xml->getAttributeValue("beatResolution");
					if( beatResolution_c != NULL ) beatResolution = atoi( beatResolution_c );
					else
					{
						//beatResolution = 960;
						std::cerr << "Missing info from file: beat resolution" << std::endl;
						return false;
					}
					
                }
				
				// ---------- view ------
                else if (!strcmp("view", xml->getNodeName()))
				{
					
					const char* xscroll_c = xml->getAttributeValue("xscroll");
					if( xscroll_c != NULL ) x_scroll_in_pixels = atoi( xscroll_c );
					else
					{
						x_scroll_in_pixels = 0;
						std::cerr << "Missing info from file: x scroll" << std::endl;
					}
					
					if(x_scroll_in_pixels < 0)
					{
						std::cerr << "Wrong x_scroll_in_pixels: " << x_scroll_in_pixels << std::endl;
						x_scroll_in_pixels = 0;
					}
					
                    const char* yscroll_c = xml->getAttributeValue("yscroll");
					if( yscroll_c != NULL ) y_scroll = atoi( yscroll_c );
					else
					{
						y_scroll = 0;
						std::cerr << "Missing info from file: y scroll" << std::endl;
					}
					
					if(y_scroll < 0)
					{
						std::cerr << "Wrong y_scroll: " << y_scroll << std::endl;
						y_scroll = 0;
					}
					
					const char* zoom_c = xml->getAttributeValue("zoom");
					if( zoom_c != NULL )
					{
						// FIXME - deprecated old zoom save, remove after some time
						if(wxString(zoom_c).Find(".")>-1)
						{
							zoom = string2double( (char*)zoom_c );
						}
						else
						{
							// correct zoom save and load method
							int zoom_i = atoi(zoom_c);
							if(zoom_i > 0 and zoom_i<501) setZoom( zoom_i );
							else return false;
						}
					}
					else
					{
						setZoom( 100 );
						std::cerr << "Missing info from file: zoom" << std::endl;
					}
					
					if(zoom <= 0)
					{
						std::cerr << "Fatal Error: Wrong Zoom: " << zoom  << "(char* = " << zoom_c << ") " << std::endl;
						setZoom( 100 );
					}
					
                }  
				
				// ---------- measure ------
				else if (!strcmp("measure", xml->getNodeName()))
				{
                    
                    const char* firstMeasure_c = xml->getAttributeValue("firstMeasure");
					if( firstMeasure_c != NULL ) firstMeasure = atoi( firstMeasure_c );
					else
					{
						std::cerr << "Missing info from file: first measure" << std::endl;
						firstMeasure = 0;
					}
					
					if(firstMeasure < 0)
					{
						std::cerr << "Wrong first measure: " << firstMeasure << std::endl;
						firstMeasure = 0;
					}
					
					const char* denom = xml->getAttributeValue("denom");
					if( denom != NULL ) measureBottomValue = atoi( denom );
					else
					{
						std::cerr << "Missing info from file: measure denom" << std::endl;
						measureBottomValue = 4;
					}
					
					if(measureBottomValue < 0)
					{
						std::cerr << "Wrong measureBottomValue: " << measureBottomValue << std::endl;
						measureBottomValue = 4;
					}
					
					const char* num = xml->getAttributeValue("num");
					if( num != NULL ) measureTopValue = atoi( num );
					else
					{
						std::cerr << "Missing info from file: measure numerator" << std::endl;
						measureTopValue = 4;
					}
					
					if(measureTopValue < 0)
					{
						std::cerr << "Wrong measureTopValue: " << measureTopValue << std::endl;
						measureTopValue = 4;
					}
					
                } 
				
				// ---------- track ------
				else if (!strcmp("track", xml->getNodeName()))
				{
					tracks.push_back( new Track(mainFrame, this) );
					if(! tracks[ tracks.size()-1 ].readFromFile(xml) )
						return false;
				}
				
				// ---------- copyright ------
				else if (!strcmp("copyright", xml->getNodeName()))
				{
					copyright_mode=true;
				}
				
				// ---------- tempo events ------
				else if (!strcmp("tempo", xml->getNodeName()))
				{
					tempo_mode=true;
				}
				// all control events in <sequence> are tempo events
                else if (!strcmp("controlevent", xml->getNodeName()))
				{
					
					if(!tempo_mode)
					{
						std::cerr << "Unexpected control event" << std::endl;
						continue;
					}
					
					int tempo_tick = -1;
					const char* tick = xml->getAttributeValue("tick");
					if( tick != NULL ) tempo_tick = atoi( tick );
					if( tempo_tick<0 )
					{
						std::cerr << "Failed to read tempo event" << std::endl;
						continue;
					}
					
					int tempo_value = -1;
					const char* value = xml->getAttributeValue("value");
					if( value != NULL ) tempo_value = atoi( value );
					if( tempo_value<0 )
					{
						std::cerr << "Failed to read tempo event" << std::endl;
						continue;
					}
					
					tempoEvents.push_back( new ControllerEvent(this, 201, tempo_tick, tempo_value) );
					
				}
				
			}// end case
				
                break;
            case irr::io::EXN_ELEMENT_END:
            {
                
                if (!strcmp("sequence", xml->getNodeName()))
				{
					goto over;
                }
				else if (!strcmp("copyright", xml->getNodeName()))
				{
					copyright_mode=false;
				}
				else if (!strcmp("tempo", xml->getNodeName()))
				{
					tempo_mode=false;
				}
				
            }
                break;
              
			default:break;
        }//end switch
    }// end while

over:
	
	mainFrame->setHorizontalScrollbar( x_scroll_in_pixels );
	glPane->render();
	
	return true;
	
}

}
